!****************************************************************************

!  Module:      multifluid_database

!  Description: These routines maintain a database of information regarding
!               multiple fluids used in hydrodynamical calculations.  This
!               information is used by hydro solvers to determine the number
!               of advection equations to solve, by equations of state and
!               source-term solvers to determine mean molecular/isotopic
!               masses, and by other modules (e.g., I/O) to output useful
!               information about a simulation.

!               "Fluids" are generally defined -- they could be test fluids
!               ("fluid A" and "fluid B" to study mixing, for example),
!               charged particles and neutrals, chemical reagents, isotopes,
!               etc.  The database doesn't care what a solver does with its
!               information, but care has been taken to make efficient and
!               flexible query methods available to user routines.  For
!               example, user routines can request information on a single
!               fluid in the database or the entire database at once.
!               Fluids can be referenced by name or by array/hash index.
!               Allocatable arrays are used instead of linked lists, and
!               routines used to search the database utilize a few core
!               routines which encapsulate the search strategy, allowing,
!               for instance, the easy replacement of a linear search with a
!               hash-table search.

!               In all of the public subroutine interfaces described below,
!               a single fluid ("f") may be specified using a string name
!               or an integer array/hash index.  A generic interface is
!               defined to determine which access routine to use.

!               Supported fluid properties and their types include:

!                       Name                    A.K.A.          Type
!                       ----                    ------          ----
!                       "name"                  name            string
!                       "short name"            chem. symbol    string
!                       "num total"             A               real
!                       "num positive"          Z               real
!                       "num neutral"           N               real
!                       "num negative"          E               real
!                       "binding energy"        Eb              real
!                       "adiabatic index"       gamma           real

!               For efficient calculations of weighted sums, the inverses
!               and squares of these quantities are also stored, although
!               they are not directly modified by the user program.

!               Numbers of charged components are stored as reals to permit
!               fractional charges and average charges/masses.  For example,
!               a user might want to follow atomic species without worrying
!               about the isotopic distribution.  Instead of defining a fluid
!               for every isotope, the user could define a fluid for each
!               element, setting A, Z, etc. to the average values for that
!               element under the assumption of some non-varying distribution
!               of isotopes.

!               Currently introspection is not supported; e.g., this module
!               does not provide routines to ask what the available properties
!               are.

!  Provides:    add_fluid_to_db (name, short_name, properties, status)
!                       A quick way to set a number of properties for an
!                       individual fluid in a single subroutine call.  Looks
!                       for the next uninitialized fluid (init_mfluid_db() sets
!                       the names of all fluids to UNINITIALIZED) and sets its
!                       properties according to the values specified in the
!                       subroutine call.  Properties can be specified in the
!                       order A, Z, N, E, ... or by keyword.  For example,

!                        call add_fluid_to_db ("helium", "He", A=4., Z=2., N=2.)

!                       Properties not specifically initialized are set to 0.
!                       The status parameter is an optional status variable.

!               set_fluid_property (f, p, v, status)
!                       Set the property p of fluid f to value v.  p is a
!                       string identifier.  The value v can be real-valued or
!                       string-valued, depending on the property being
!                       modified.  status is an optional variable which will
!                       be set to 0 if the operation was successful, -1 if not.
!                       Reasons why the operation can fail include:  f out
!                       of bounds if f is an index; f not found if f is a
!                       string name; p is not a valid property identifier;
!                       v is not a valid value for the given property.

!               get_fluid_property (f, p, v, status)
!                       Like the setting version, but v now receives the
!                       value instead of setting it.

!               get_mfluid_property (p, v, status)
!                       Return the value of the property p for all defined
!                       fluids.  v must be an array of the correct type.
!                       status is an optional exit status variable.  v is
!                       filled with values up to the minimum of (its size,
!                       number of fluids).

!               n_fluids
!                       The number of fluids in the multifluid database.

!               find_fluid_index (f, i)
!                       Find the database index i of a fluid named f.
!                       If init_mfluid_db() has not been called, terminates with
!                       an error.  If the fluid name is not found, terminates
!                       with an error.  Errors are signaled by setting i to
!                       mfluid_status_fail.

!               query_mfluid_sum (p, w, v, status)
!                       Given a property name p and an array of weights w,
!                       return the weighted sum of the chosen property in v.
!                       w should be an array of length equal to the number of
!                       fluids in the database, or else a one-element array.
!                       If it is neither, or if the named property is invalid,
!                       the routine terminates.  The optional status variable
!                       is set to mfluid_status_ok or mfluid_status_fail
!                       depending on whether the summing operation was
!                       successful.

!               query_mfluid_suminv (p, w, v, status)
!                       Same as query_mfluid_sum(), but compute the weighted
!                       sum of the inverse of the chosen property.

!               query_mfluid_sumfrc (p, w, v, status)
!                       Same as query_mfluid_sum(), but compute the weighted
!                       sum of the chosen property divided by the total number
!                       of particles (A).

!               query_mfluid_sumsqr (p, w, v, status)
!                       Same as query_mfluid_sum(), but compute the weighted
!                       sum of the square of the chosen property.

!               init_mfluid_db ()
!                       Initialize the multifluid database.  If this has not
!                       been called and one of the other routines is called,
!                       that routine terminates with an error.

!               list_mfluid_db (lun)
!                       List the contents of the multifluid database in a
!                       snappy table format.  Output goes to the logical I/O
!                       unit indicated by the lun parameter.


        module multifluid_database

!****************************************************************************

!               Public data types, constants, and routines.

        implicit none
        private
        public :: set_fluid_property, get_fluid_property, 
     &            get_mfluid_property,
     &            init_mfluid_db, list_mfluid_db, add_fluid_to_db,
     &            query_mfluid_sum, n_fluids, find_fluid_index,
     &            query_mfluid_suminv, query_mfluid_sumsqr,
     &            query_mfluid_sumfrc
        public :: mfluid_status_ok, mfluid_status_fail
        public :: mf_prop_gamma, mf_prop_A, mf_prop_Z, mf_prop_N
        public :: mf_prop_E    , mf_prop_A, mf_prop_Eb

        save

!----------------------------------------------------------------------------

!               Constants used to return status information.

        integer, parameter :: mfluid_status_ok      = 0,
     &                        mfluid_status_fail    = -1,
     &                        mfluid_invalid        = -2,
     &                        type_string           = 1,
     &                        type_real             = 2

!               Constant and derived type declarations for properties.

        integer, parameter      :: n_str_props  = 2
        integer, parameter      :: n_real_props = 6
        integer, parameter      :: n_props = n_str_props + n_real_props

!               For each property, we store its name, its type (string or
!               real), and its index in the appropriate array.  For example,
!               the value of the parameter "short name" is stored in each
!               fluid's str_props(2).  Thus the index of the "short name"
!               parameter is 2 (and its type is type_string).

        integer :: mf_prop_gamma, mf_prop_A, mf_prop_Z, mf_prop_N
        integer :: mf_prop_E    , mf_prop_A, mf_prop_Eb

        type property_type
          character(len=80)     :: name
          integer               :: type, index
        end type

        type (property_type)    :: mfluid_props(n_props) =
     &       (/ property_type("name",                type_string, 1),
     &          property_type("short name",          type_string, 2),
     &          property_type("num total",           type_real,   1),
     &          property_type("num positive",        type_real,   2),
     &          property_type("num neutral",         type_real,   3),
     &          property_type("num negative",        type_real,   4),
     &          property_type("binding energy",      type_real,   5),
     &          property_type("adiabatic index",     type_real,   6) /)

!               Database variables.

!               For each fluid, we store real-valued properties (e.g., A)
!               and string-valued properties (e.g., name) in separate arrays.
!               We also store (and keep up-to-date whenever set_* is called)
!               the inverses and squares of real-valued properties.

        type fluid_type
          character(len=80)             :: str_props(n_str_props)
          real                          :: real_props(n_real_props)
          real                          :: real_props_inv(n_real_props)
          real                          :: real_props_sqr(n_real_props)
          real                          :: real_props_frc(n_real_props)
        end type

        integer                         :: n_fluids
        type (fluid_type), allocatable  :: mfluid_db(:)


        logical                         :: init_was_called = .false.

!----------------------------------------------------------------------------

!               Generic subroutine interfaces.  Routines must take indices
!               or string names as fluid identifiers, and they must set or
!               retrieve string-valued or real-valued properties.

        interface set_fluid_property
          module procedure set_str_prop_by_str
          module procedure set_str_prop_by_ind
          module procedure set_num_prop_by_str
          module procedure set_num_prop_by_ind
          module procedure set_str_iprop_by_str
          module procedure set_str_iprop_by_ind
          module procedure set_num_iprop_by_str
          module procedure set_num_iprop_by_ind
        end interface

        interface get_fluid_property
          module procedure get_str_prop_by_str
          module procedure get_str_prop_by_ind
          module procedure get_num_prop_by_str
          module procedure get_num_prop_by_ind
          module procedure get_str_iprop_by_str
          module procedure get_str_iprop_by_ind
          module procedure get_num_iprop_by_str
          module procedure get_num_iprop_by_ind
        end interface

        interface get_mfluid_property
          module procedure get_all_str_prop
          module procedure get_all_num_prop
          module procedure get_all_str_iprop
          module procedure get_all_num_iprop
        end interface


        interface query_mfluid_sum
          module procedure query_mfluid_sum_prop
          module procedure query_mfluid_sum_iprop
        end interface

        interface query_mfluid_suminv
          module procedure query_mfluid_suminv_prop
          module procedure query_mfluid_suminv_iprop
        end interface

        interface query_mfluid_sumfrc
          module procedure query_mfluid_sumfrc_prop
          module procedure query_mfluid_sumfrc_iprop
        end interface

        interface query_mfluid_sumsqr
          module procedure query_mfluid_sumsqr_prop
C          module procedure query_mfluid_sumsqr_iprop
        end interface

!----------------------------------------------------------------------------

        contains

!****************************************************************************

!  Routine:     init_mfluid_db ()

!  Description: Initialize the multifluid database.  If this has not
!               been called and one of the other routines is called,
!               that routine terminates with an error.

!  Arguments:   max_fluids              Number of fluids to allocate space
!                                         for.
!               status                  An optional integer variable to
!                                         receive the exit status.

        subroutine init_mfluid_db (max_fluids, status)

!----------------------------------------------------------------------------

        integer, intent(in)             :: max_fluids
        integer, optional, intent(out)  :: status

        integer                         :: istat, i

!----------------------------------------------------------------------------

!               Unless we make it through, we've failed.

        if (present(status)) status = mfluid_status_fail

!               If we're already initialized, don't do it again.

        if (init_was_called) then
          write (*,*) 'init_mfluid_db:  already initialized'
          return
        endif

!               Allocate storage for the multifluid database.
!               Set the names of uninitialized fluids to UNINITIALIZED.

        n_fluids = max_fluids
        allocate (mfluid_db(n_fluids), stat=istat)
        if (istat /= 0) return
        do i = 1, n_fluids
          mfluid_db(i)%str_props(1)      = "UNINITIALIZED"
          mfluid_db(i)%str_props(2)      = "UN"
          mfluid_db(i)%real_props(:)     = 0.
          mfluid_db(i)%real_props_inv(:) = 0.
          mfluid_db(i)%real_props_sqr(:) = 0.
          mfluid_db(i)%real_props_frc(:) = 0.
        enddo

!               init_mfluid_db has been called successfully.

        init_was_called = .true.
        if (present(status)) status = mfluid_status_ok

!
!        define the property numbers
!
        call check_prop_name("num total",      type_real,mf_prop_A)
        call check_prop_name("num positive",   type_real,mf_prop_Z)
        call check_prop_name("num neutral",    type_real,mf_prop_N)
        call check_prop_name("num negative",   type_real,mf_prop_E)
        call check_prop_name("binding energy", type_real,mf_prop_Eb)
        call check_prop_name("adiabatic index",type_real,mf_prop_gamma)

!----------------------------------------------------------------------------

        return
        end subroutine init_mfluid_db

!****************************************************************************

!  Routine:     add_fluid_to_db (name, short_name, properties, status)

!  Description: A quick way to set a number of properties for an
!               individual fluid in a single subroutine call.  Looks
!               for the next uninitialized fluid (init_mfluid_db() sets
!               the names of all fluids to UNINITIALIZED) and sets its
!               properties according to the values specified in the
!               subroutine call.  Properties can be specified in the
!               order A, Z, N, E, ... or by keyword.  For example,

!                call add_fluid_to_db ("helium", "He", A=4., Z=2., N=2.)

!               Properties not specifically initialized are set to 0.
!               The status parameter is an optional status variable name.


        subroutine add_fluid_to_db (name, short_name,
     &                              A, Z, N, E, Eb, gamma, status)

!----------------------------------------------------------------------------

        character(len=*)                :: name, short_name
        real, optional                  :: A, Z, N, E, Eb, gamma
        integer, optional               :: status

        integer                         :: index

!----------------------------------------------------------------------------

!               Get the index of the first fluid named UNINITIALIZED.

        call find_fluid_index ("UNINITIALIZED", index)

        if (index == mfluid_invalid) then
          write (*,*) "add_fluid_to_db:  no uninitialized fluids ",
     &                  "found in db"
          if (present(status)) status = mfluid_status_fail
          return
        endif

!               Now set its name and other properties.

        call set_fluid_property (index, "name", name)
        call set_fluid_property (index, "short name", short_name)

        if (present(A))
     &    call set_fluid_property (index, "num total", A)
        if (present(Z))
     &    call set_fluid_property (index, "num positive", Z)
        if (present(N))
     &    call set_fluid_property (index, "num neutral", N)
        if (present(E))
     &    call set_fluid_property (index, "num negative", E)
        if (present(Eb))
     &    call set_fluid_property (index, "binding energy", Eb)
        if (present(gamma))
     &    call set_fluid_property (index, "adiabatic index", gamma)

!----------------------------------------------------------------------------

        return
        end subroutine add_fluid_to_db

!****************************************************************************

!  Routine:     set_fluid_property (f, p, v, status)

!  Description: Set the property p of fluid f to value v.  p is a
!               string identifier.  The value v can be real-valued or
!               string-valued, depending on the property being
!               modified.  status is an optional variable which will
!               be set to 0 if the operation was successful, -1 if not.
!               Reasons why the operation can fail include:  f out
!               of bounds if f is an index; f not found if f is a
!               string name; p is not a valid property identifier;
!               v is not a valid value for the given property.

!============================================================================

!
!   this is the lowest level routine; it does the copying.  
!   the higher-up levels have to do the input checking.
!
        subroutine set_str_prop (ifluid, iprop, value)

        integer                         :: ifluid, iprop
        character(len=*)                :: value

        mfluid_db(ifluid)%str_props(iprop) = value

        return
        end subroutine set_str_prop

!----------------------------------------------------------------------------



        subroutine set_str_prop_by_str (fluid, prop, value, status)

!----------------------------------------------------------------------------

        character(len=*)                :: fluid, prop, value
        integer, optional               :: status

        integer                         :: ifluid, iprop

!----------------------------------------------------------------------------

        call find_fluid_index (fluid, ifluid)

        if (ifluid == mfluid_invalid) then
          write (*,*) "set_fluid_property:  fluid name '",
     &                fluid(1:len_trim(fluid)), "' not found in db"
          if (present(status)) status = mfluid_status_fail
          return
        endif

        call check_prop_name (prop, type_string, iprop)
        if (iprop == mfluid_status_fail) then
          write (*,*) "set_fluid_property:  string-valued property",
     &                  " name '", prop(1:len_trim(prop)),
     &                  "' not recognized"
          status = mfluid_status_fail
          return
        endif

        call set_str_prop (ifluid, iprop, value)

!----------------------------------------------------------------------------

        return
        end subroutine set_str_prop_by_str

!============================================================================

        subroutine set_str_prop_by_ind (ifluid, prop, value, status)

!----------------------------------------------------------------------------

        integer                         :: ifluid, iprop
        character(len=*)                :: prop, value
        integer, optional               :: status

        integer                         :: istat

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the fluid index and the property name to make sure
!               they're valid.  If the property name is valid, check_prop_name
!               returns the property's index.

        call check_fluid_index (ifluid, istat)
        if (istat == mfluid_status_fail) then
          write (*,*) "set_fluid_property:  fluid index ", ifluid,
     &                  " not found in db"
          return
        endif

        call check_prop_name (prop, type_string, iprop)
        if (iprop == mfluid_status_fail) then
          write (*,*) "set_fluid_property:  string-valued property",
     &                  " name '", prop(1:len_trim(prop)),
     &                  "' not recognized"
          return
        endif

!               Everything's OK; go ahead and set the property.

        call set_str_prop (ifluid, iprop, value)

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine set_str_prop_by_ind

!----------------------------------------------------------------------------



        subroutine set_str_iprop_by_str (fluid, iprop, value, status)

!----------------------------------------------------------------------------

        character(len=*)                :: fluid, value
        integer, optional               :: status

        integer                         :: ifluid, iprop

!----------------------------------------------------------------------------

        call find_fluid_index (fluid, ifluid)

        if (ifluid == mfluid_invalid) then
          write (*,*) "set_fluid_property:  fluid name '",
     &                fluid(1:len_trim(fluid)), "' not found in db"
          if (present(status)) status = mfluid_status_fail
          return
        endif

        if (iprop .lt. 1 .or. iprop .gt. n_str_props) then
          write (*,*) "set_fluid_property:  invalid property",iprop
          status = mfluid_status_fail
          return
        endif

        call set_str_prop (ifluid, iprop, value)

!----------------------------------------------------------------------------

        return
        end subroutine set_str_iprop_by_str

!============================================================================

        subroutine set_str_iprop_by_ind (fluid, iprop, value, status)

!----------------------------------------------------------------------------

        integer                         :: fluid
        character(len=*)                :: value
        integer, optional               :: status

        integer                         :: istat, iprop

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the fluid index and the property name to make sure
!               they're valid.  If the property name is valid, check_prop_name
!               returns the property's index.

        call check_fluid_index (fluid, istat)
        if (istat == mfluid_status_fail) then
          write (*,*) "set_fluid_property:  fluid index ", fluid,
     &                  " not found in db"
          return
        endif

        if (iprop .lt. 1 .or. iprop .gt. n_str_props) then
          write (*,*) "set_fluid_property:  invalid str property",iprop
          status = mfluid_status_fail
          return
        endif

!               Everything's OK; go ahead and set the property.

        call set_str_prop (fluid, iprop, value)

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine set_str_iprop_by_ind






!============================================================================

        subroutine set_num_prop_by_str (fluid, prop, value, status)

!----------------------------------------------------------------------------

        character(len=*)                :: fluid, prop
        real                            :: value
        integer, optional               :: status

        integer                         :: index

!----------------------------------------------------------------------------

        call find_fluid_index (fluid, index)

        if (index == mfluid_invalid) then
          write (*,*) "set_fluid_property:  fluid name '",
     &                fluid(1:len_trim(fluid)), "' not found in db"
          if (present(status)) status = mfluid_status_fail
          return
        endif

        call set_num_prop_by_ind (index, prop, value, status)

!----------------------------------------------------------------------------

        return
        end subroutine set_num_prop_by_str

!============================================================================

        subroutine set_num_prop_by_ind (fluid, prop, value, status)

!----------------------------------------------------------------------------

        integer                         :: fluid
        character(len=*)                :: prop
        real                            :: value
        integer, optional               :: status

        integer                         :: istat, i
        real                            :: total

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the fluid index and the property name to make sure
!               they're valid.  If the property name is valid, check_prop_name
!               returns the property's index.

        call check_fluid_index (fluid, istat)
        if (istat == mfluid_status_fail) then
          write (*,*) "set_fluid_property:  fluid index ", fluid,
     &                  " not found in db"
          return
        endif

        call check_prop_name (prop, type_real, istat)
        if (istat == mfluid_status_fail) then
          write (*,*) "set_fluid_property:  real-valued property",
     &                  " name '", prop(1:len_trim(prop)),
     &                  "' not recognized"
          return
        endif

!               Everything's OK; go ahead and set the property.

        mfluid_db(fluid)%real_props(istat)      = value
        mfluid_db(fluid)%real_props_inv(istat)  = 1.d0 / value
        mfluid_db(fluid)%real_props_sqr(istat)  = value ** 2

!               Also store the ratio of each property to the total number
!               of particles for the fluid.

        call get_fluid_property (fluid, "num total", total)
        do i = 1, n_real_props
          mfluid_db(fluid)%real_props_frc(i) =
     &          mfluid_db(fluid)%real_props(i) / total
        enddo

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine set_num_prop_by_ind


!============================================================================

        subroutine set_num_iprop_by_str (fluid, iprop, value, status)

!----------------------------------------------------------------------------

        character(len=*)                :: fluid
        real                            :: value
        integer, optional               :: status

        integer                         :: index, iprop

!----------------------------------------------------------------------------

        call find_fluid_index (fluid, index)

        if (index == mfluid_invalid) then
          write (*,*) "set_fluid_property:  fluid name '",
     &                fluid(1:len_trim(fluid)), "' not found in db"
          if (present(status)) status = mfluid_status_fail
          return
        endif

        call set_num_iprop_by_ind (index, iprop, value, status)

!----------------------------------------------------------------------------

        return
        end subroutine set_num_iprop_by_str

!============================================================================

        subroutine set_num_iprop_by_ind (fluid, iprop, value, status)

!----------------------------------------------------------------------------

        integer                         :: fluid
        real                            :: value
        integer, optional               :: status

        integer                         :: istat, i, iprop
        real                            :: total

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the fluid index and the property name to make sure
!               they're valid.  If the property name is valid, check_prop_name
!               returns the property's index.

        call check_fluid_index (fluid, istat)
        if (istat == mfluid_status_fail) then
          write (*,*) "set_fluid_property:  fluid index ", fluid,
     &                  " not found in db"
          return
        endif

        if (iprop .lt. 1 .or. iprop .gt. n_real_props) then
          write(*,*) "set_fluid_property: invalid real property ",iprop
          return
        endif

!               Everything's OK; go ahead and set the property.

        mfluid_db(fluid)%real_props(iprop)      = value
        mfluid_db(fluid)%real_props_inv(iprop)  = 1.d0 / value
        mfluid_db(fluid)%real_props_sqr(iprop)  = value ** 2

!               Also store the ratio of each property to the total number
!               of particles for the fluid.

        call get_fluid_property (fluid, mf_prop_A, total)
        do i = 1, n_real_props
          mfluid_db(fluid)%real_props_frc(i) =
     &          mfluid_db(fluid)%real_props(i) / total
        enddo

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine set_num_iprop_by_ind




!****************************************************************************

!  Routine:     get_fluid_property (f, p, v, status)

!  Description: Return the value of the property p of fluid f in variable
!               v.  p is a string identifier.  The variable v can be real
!               or string, depending on the property being queried.  status
!               is an optional variable which will be set to 0 if the
!               operation was successful, -1 if not.

!============================================================================


        subroutine get_str_prop_by_str (fluid, prop, variable, status)

!----------------------------------------------------------------------------

        character(len=*)                :: fluid, prop, variable
        integer, optional               :: status

        integer                         :: index

!----------------------------------------------------------------------------

        call find_fluid_index (fluid, index)

        if (index == mfluid_invalid) then
          write (*,*) "get_fluid_property:  fluid name '",
     &                fluid(1:len_trim(fluid)), "' not found in db"
          if (present(status)) status = mfluid_status_fail
          return
        endif

        call get_str_prop_by_ind (index, prop, variable, status)

!----------------------------------------------------------------------------

        return
        end subroutine get_str_prop_by_str

!============================================================================

        subroutine get_str_prop_by_ind (fluid, prop, variable, status)

!----------------------------------------------------------------------------

        integer                         :: fluid
        character(len=*)                :: prop, variable
        integer, optional               :: status

        integer                         :: istat

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the fluid index and the property name to make sure
!               they're valid.  If the property name is valid, check_prop_name
!               returns the property's index.

        call check_fluid_index (fluid, istat)
        if (istat == mfluid_status_fail) then
          write (*,*) "get_fluid_property:  fluid index ", fluid,
     &                  " not found in db"
          return
        endif

        call check_prop_name (prop, type_string, istat)
        if (istat == mfluid_status_fail) then
          write (*,*) "get_fluid_property:  string-valued property",
     &                  " name '", prop(1:len_trim(prop)),
     &                  "' not recognized"
          return
        endif

!               Everything's OK; go ahead and return the property.

        variable = mfluid_db(fluid)%str_props(istat)

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine get_str_prop_by_ind

!============================================================================

        subroutine get_num_prop_by_str (fluid, prop, variable, status)

!----------------------------------------------------------------------------

        character(len=*)                :: fluid, prop
        real                            :: variable
        integer, optional               :: status

        integer                         :: index

!----------------------------------------------------------------------------

        call find_fluid_index (fluid, index)

        if (index == mfluid_invalid) then
          write (*,*) "get_fluid_property:  fluid name '",
     &                fluid(1:len_trim(fluid)), "' not found in db"
          if (present(status)) status = mfluid_status_fail
          return
        endif

        call get_num_prop_by_ind (index, prop, variable, status)

!----------------------------------------------------------------------------

        return
        end subroutine get_num_prop_by_str

!============================================================================

        subroutine get_num_prop_by_ind (fluid, prop, variable, status)

!----------------------------------------------------------------------------

        integer                         :: fluid
        character(len=*)                :: prop
        real                            :: variable
        integer, optional               :: status

        integer                         :: istat

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the fluid index and the property name to make sure
!               they're valid.  If the property name is valid, check_prop_name
!               returns the property's index.

        call check_fluid_index (fluid, istat)
        if (istat == mfluid_status_fail) then
          write (*,*) "set_fluid_property:  fluid index ", fluid,
     &                  " not found in db"
          return
        endif

        call check_prop_name (prop, type_real, istat)
        if (istat == mfluid_status_fail) then
          write (*,*) "set_fluid_property:  real-valued property",
     &                  " name '", prop(1:len_trim(prop)),
     &                  "' not recognized"
          return
        endif

!               Everything's OK; go ahead and return the property.

        variable = mfluid_db(fluid)%real_props(istat)

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine get_num_prop_by_ind



        subroutine get_str_iprop_by_str (fluid, iprop, variable, status)

!----------------------------------------------------------------------------

        character(len=*)                :: fluid, variable
        integer, optional               :: status

        integer                         :: index, iprop

!----------------------------------------------------------------------------

        call find_fluid_index (fluid, index)

        if (index == mfluid_invalid) then
          write (*,*) "get_fluid_property:  fluid name '",
     &                fluid(1:len_trim(fluid)), "' not found in db"
          if (present(status)) status = mfluid_status_fail
          return
        endif

        call get_str_iprop_by_ind (index, iprop, variable, status)

!----------------------------------------------------------------------------

        return
        end subroutine get_str_iprop_by_str

!============================================================================

        subroutine get_str_iprop_by_ind (fluid, iprop, variable, status)

!----------------------------------------------------------------------------

        integer                         :: fluid
        character(len=*)                :: variable
        integer, optional               :: status

        integer                         :: istat, iprop

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the fluid index and the property name to make sure
!               they're valid.  If the property name is valid, check_prop_name
!               returns the property's index.

        call check_fluid_index (fluid, istat)
        if (istat == mfluid_status_fail) then
          write (*,*) "get_fluid_property:  fluid index ", fluid,
     &                  " not found in db"
          return
        endif

        if (iprop .lt. 1 .or. iprop .gt. n_str_props) then 
          write(*,*) "get_fluid_property:  invalid str property ",iprop
          return
        endif

!               Everything's OK; go ahead and return the property.

        variable = mfluid_db(fluid)%str_props(istat)

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine get_str_iprop_by_ind

!============================================================================

        subroutine get_num_iprop_by_str (fluid, iprop, variable, status)

!----------------------------------------------------------------------------

        character(len=*)                :: fluid
        real                            :: variable
        integer, optional               :: status

        integer                         :: index, iprop

!----------------------------------------------------------------------------

        call find_fluid_index (fluid, index)

        if (index == mfluid_invalid) then
          write (*,*) "get_fluid_property:  fluid name '",
     &                fluid(1:len_trim(fluid)), "' not found in db"
          if (present(status)) status = mfluid_status_fail
          return
        endif

        call get_num_iprop_by_ind (index, iprop, variable, status)

!----------------------------------------------------------------------------

        return
        end subroutine get_num_iprop_by_str

!============================================================================

        subroutine get_num_iprop_by_ind (fluid, iprop, variable, status)

!----------------------------------------------------------------------------

        integer                         :: fluid
        integer                         :: iprop
        real                            :: variable
        integer, optional               :: status

        integer                         :: istat

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the fluid index and the property name to make sure
!               they're valid.  If the property name is valid, check_prop_name
!               returns the property's index.

        call check_fluid_index (fluid, istat)
        if (istat == mfluid_status_fail) then
          write (*,*) "set_fluid_property:  fluid index ", fluid,
     &                  " not found in db"
          return
        endif

        if (iprop .lt. 1 .or. iprop .gt. n_real_props) then
          write(*,*) "set_fluid_property: invalid real property ",iprop
          return
        endif

!               Everything's OK; go ahead and return the property.

        variable = mfluid_db(fluid)%real_props(iprop)

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine get_num_iprop_by_ind

!****************************************************************************

!  Routine:     get_mfluid_property (p, v, status)

!  Description: Return the value of the property p for all defined fluids.
!               v must be an array of the correct type.  status
!               is an optional exit status variable.  v is filled with values
!               up to the minimum of (its size, number of fluids).

!============================================================================

        subroutine get_all_str_prop (prop, value, status)

!----------------------------------------------------------------------------

        character(len=*)                :: prop
        character(len=*)                :: value(:)
        integer, optional               :: status

        integer                         :: istat, index, nvals

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the supplied property identifier for validity.
!               If it's OK, return the index of the property to retrieve.

        call check_prop_name (prop, type_string, index)
        if (index == mfluid_status_fail) then
          write (*,*) "get_mfluid_property:  string-valued property",
     &                  " name '", prop(1:len_trim(prop)),
     &                  "' not recognized"
          return
        endif

!               Determine how many results we can return.

        nvals = min( n_fluids, size(value, 1) )

!               Now fill the array.

        value = mfluid_db(:nvals)%str_props(index)

!               Success.

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine get_all_str_prop

!============================================================================

        subroutine get_all_num_prop (prop, value, status)

!----------------------------------------------------------------------------

        character(len=*)                :: prop
        real                            :: value(:)
        integer, optional               :: status

        integer                         :: istat, index, nvals

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the supplied property identifier for validity.
!               If it's OK, return the index of the property to retrieve.

        call check_prop_name (prop, type_real, index)
        if (index == mfluid_status_fail) then
          write (*,*) "get_mfluid_property:  real-valued property",
     &                  " name '", prop(1:len_trim(prop)),
     &                  "' not recognized"
          return
        endif

!               Determine how many results we can return.

        nvals = min( n_fluids, size(value, 1) )

!               Now fill the array.

        value = mfluid_db(:nvals)%real_props(index)

!               Success.

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine get_all_num_prop



        subroutine get_all_str_iprop (iprop, value, status)

!----------------------------------------------------------------------------

        integer                         :: iprop
        character(len=*)                :: value(:)
        integer, optional               :: status

        integer                         :: istat, index, nvals

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the supplied property identifier for validity.
!               If it's OK, return the index of the property to retrieve.

        index = iprop
        if (iprop .lt. 1 .or. iprop .gt. n_str_props) then
          write (*,*) "get_mfluid_property:  string-valued property ",
     &                  iprop," not recognized"
          return
        endif

!               Determine how many results we can return.

        nvals = min( n_fluids, size(value, 1) )

!               Now fill the array.

        value = mfluid_db(:nvals)%str_props(index)

!               Success.

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine get_all_str_iprop

!============================================================================

        subroutine get_all_num_iprop (iprop, value, status)

!----------------------------------------------------------------------------

        integer                         :: iprop
        real                            :: value(:)
        integer, optional               :: status

        integer                         :: istat, index, nvals

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the supplied property identifier for validity.
!               If it's OK, return the index of the property to retrieve.

        index = iprop
        if (iprop .lt. 1 .or. iprop .gt. n_str_props) then
          write (*,*) "get_mfluid_property:  string-valued property",
     &                  iprop," not recognized"
          return
        endif

!               Determine how many results we can return.

        nvals = min( n_fluids, size(value, 1) )

!               Now fill the array.

        value = mfluid_db(:nvals)%real_props(index)

!               Success.

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine get_all_num_iprop

!****************************************************************************

!  Routine:     query_mfluid_sum_iprop (iprop, weights, variable, status)

!  Description: Given a property name and an array of weights, compute the
!               weighted sum of the chosen property.  weights should be
!               an array of length equal to the number of fluids in the
!               database, or else a one-element array.  If it is neither,
!               or if the named property is invalid, the routine terminates.
!               The optional status variable is set to mfluid_status_ok or
!               mfluid_status_fail depending on whether the summing
!               operation was successful.  If successful, the result is
!               returned in variable.


        subroutine query_mfluid_sum_iprop(iprop,weights,
     +                                   variable,status)

!----------------------------------------------------------------------------

        integer                         :: iprop
        real                            :: weights(:), variable
        integer, optional               :: status

        integer                         :: istat, index, nvals

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the supplied property identifier for validity.
!               If it's OK, return the index of the property to retrieve.

        if (iprop .lt. 1 .or. iprop .gt. n_real_props) then
          write (*,*) "query_mfluid_sum:  real-valued property ",
     &                  iprop," not recognized"
          return
        endif

!               Determine whether weights makes sense.

        nvals = size(weights, 1)
        if ((nvals /= 1) .and. (nvals /= n_fluids)) then
          write (*,*) "query_mfluid_sum:  invalid weight array"
          return
        endif

!               Now compute the sum.

        if (nvals == 1) then

          variable = weights(1) * sum( mfluid_db%real_props(iprop) )

        else

          variable = sum( weights * mfluid_db%real_props(iprop) )

        endif

!               Success.

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine query_mfluid_sum_iprop

!****************************************************************************

!  Routine:     query_mfluid_sum_prop (prop, weights, variable, status)

!  Description: Given a property name and an array of weights, compute the
!               weighted sum of the chosen property.  weights should be
!               an array of length equal to the number of fluids in the
!               database, or else a one-element array.  If it is neither,
!               or if the named property is invalid, the routine terminates.
!               The optional status variable is set to mfluid_status_ok or
!               mfluid_status_fail depending on whether the summing
!               operation was successful.  If successful, the result is
!               returned in variable.


        subroutine query_mfluid_sum_prop(prop,weights,variable,status)

!----------------------------------------------------------------------------

        character(len=*)                :: prop
        real                            :: weights(:), variable
        integer, optional               :: status

        integer                         :: istat, index, nvals

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the supplied property identifier for validity.
!               If it's OK, return the index of the property to retrieve.

        call check_prop_name (prop, type_real, index)
        if (index == mfluid_status_fail) then
          write (*,*) "query_mfluid_sum:  real-valued property",
     &                  " name '", prop(1:len_trim(prop)),
     &                  "' not recognized"
          return
        endif

        call query_mfluid_sum_iprop(index,weights,variable,status)

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------
        return
        end subroutine query_mfluid_sum_prop

!****************************************************************************

!  Routine:     query_mfluid_suminv (prop, weights, variable, status)

!  Description: Like query_mfluid_sum, but sums the inverses of the chosen
!               property.


       subroutine query_mfluid_suminv_prop(prop,weights,variable,status)

!----------------------------------------------------------------------------

        character(len=*)                :: prop
        real                            :: weights(:), variable
        integer, optional               :: status

        integer                         :: istat, index, nvals

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the supplied property identifier for validity.
!               If it's OK, return the index of the property to retrieve.

        call check_prop_name (prop, type_real, index)
        if (index == mfluid_status_fail) then
          write (*,*) "query_mfluid_suminv:  real-valued property",
     &                  " name '", prop(1:len_trim(prop)),
     &                  "' not recognized"
          return
        endif

        call query_mfluid_suminv_iprop(index,weights,variable,status)

!----------------------------------------------------------------------------

        return
        end subroutine query_mfluid_suminv_prop



       subroutine query_mfluid_suminv_iprop(iprop,weights,
     +                                       variable,status)

!----------------------------------------------------------------------------

        integer                         :: iprop
        real                            :: weights(:), variable
        integer, optional               :: status

        integer                         :: istat, index, nvals

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the supplied property identifier for validity.
!               If it's OK, return the index of the property to retrieve.

        if (iprop .lt. 1 .or. iprop .gt. n_real_props) then
          write (*,*) "query_mfluid_suminv:  real-valued property ",
     &                  iprop," not recognized"
          return
        endif

!               Determine whether weights makes sense.

        nvals = size(weights, 1)
        if ((nvals /= 1) .and. (nvals /= n_fluids)) then
          write (*,*) "query_mfluid_suminv:  invalid weight array"
          return
        endif

!               Now compute the sum.

        if (nvals == 1) then

          variable = weights(1) * sum( mfluid_db%real_props_inv(iprop) )

        else

          variable = sum( weights * mfluid_db%real_props_inv(iprop) )

        endif

!               Success.

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine query_mfluid_suminv_iprop

!****************************************************************************

!  Routine:     query_mfluid_sumfrc (prop, weights, variable, status)

!  Description: Like query_mfluid_sum, but sums the ratio of the chosen
!               property to the total number of particles in the fluid (A).


      subroutine query_mfluid_sumfrc_prop(prop,weights,variable,status)

!----------------------------------------------------------------------------

        character(len=*)                :: prop
        real                            :: weights(:), variable
        integer, optional               :: status

        integer                         :: istat, index, nvals

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the supplied property identifier for validity.
!               If it's OK, return the index of the property to retrieve.

        call check_prop_name (prop, type_real, index)
        if (index == mfluid_status_fail) then
          write (*,*) "query_mfluid_sumfrc:  real-valued property",
     &                  " name '", prop(1:len_trim(prop)),
     &                  "' not recognized"
          return
        endif

        call query_mfluid_sumfrc_iprop(index,weights,variable,status)
!----------------------------------------------------------------------------

        return
        end subroutine query_mfluid_sumfrc_prop


      subroutine query_mfluid_sumfrc_iprop(iprop,weights,
     +                                        variable,status)

!----------------------------------------------------------------------------

        integer                         :: iprop
        real                            :: weights(:), variable
        integer, optional               :: status

        integer                         :: istat, index, nvals

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the supplied property identifier for validity.
!               If it's OK, return the index of the property to retrieve.

        if (iprop .lt. 1 .or. iprop .gt. n_real_props) then
          write (*,*) "query_mfluid_sumfrc:  real-valued property ",
     &                  iprop," not recognized"
          return
        endif

!               Determine whether weights makes sense.

        nvals = size(weights, 1)
        if ((nvals /= 1) .and. (nvals /= n_fluids)) then
          write (*,*) "query_mfluid_sumfrc:  invalid weight array"
          return
        endif

!               Now compute the sum.

        if (nvals == 1) then

          variable = weights(1) * sum( mfluid_db%real_props_frc(iprop) )

        else

          variable = sum( weights * mfluid_db%real_props_frc(iprop) )

        endif

!               Success.

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine query_mfluid_sumfrc_iprop

!****************************************************************************

!  Routine:     query_mfluid_sumsqr (prop, weights, variable, status)

!  Description: Like query_mfluid_sum, but sums the squares of the chosen
!               property.


       subroutine query_mfluid_sumsqr_prop(prop,weights,variable,status)

!----------------------------------------------------------------------------

        character(len=*)                :: prop
        real                            :: weights(:), variable
        integer, optional               :: status

        integer                         :: istat, index, nvals

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        if (present(status)) status = mfluid_status_fail

!               Check the supplied property identifier for validity.
!               If it's OK, return the index of the property to retrieve.

        call check_prop_name (prop, type_real, index)
        if (index == mfluid_status_fail) then
          write (*,*) "query_mfluid_sumsqr:  real-valued property",
     &                  " name '", prop(1:len_trim(prop)),
     &                  "' not recognized"
          return
        endif

!               Determine whether weights makes sense.

        nvals = size(weights, 1)
        if ((nvals /= 1) .and. (nvals /= n_fluids)) then
          write (*,*) "query_mfluid_sumsqr:  invalid weight array"
          return
        endif

!               Now compute the sum.

        if (nvals == 1) then

          variable = weights(1) * sum( mfluid_db%real_props_sqr(index) )

        else

          variable = sum( weights * mfluid_db%real_props_sqr(index) )

        endif

!               Success.

        if (present(status)) status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine query_mfluid_sumsqr_prop

!****************************************************************************

!  Routine:     find_fluid_index (fluid, index)

!  Description: Find the database index of a fluid with a given name.
!               If init_mfluid_db() has not been called, terminate with
!               an error.  If the fluid name is not found, terminate with
!               an error.  Errors are signaled by setting index to
!               mfluid_status_fail.


        subroutine find_fluid_index (fluid, index)

!----------------------------------------------------------------------------

        character(len=*)                :: fluid
        integer                         :: index

        integer                         :: i

!----------------------------------------------------------------------------

!               Fail unless we succeed.

        index = mfluid_status_fail

!               Has init_mfluid_db() been called?

        if (.not. init_was_called) then
          write (*,*) 'find_fluid_index:  must call init_mfluid_db()',
     &                  ' first'
          return
        endif

!               Here's where we would set index to a hash value if we were
!               doing hash-table lookup.  Currently we just search the
!               array.

!               Warning:  some external routines might call this one and
!               use its result to interpret the ordering of block-structured
!               grid values of multiple fluid densities.  In that case,
!               the interpretation of the index as a straightforward ordering
!               will have to be retained.

!               We also assume here that the first string property of each
!               fluid is its name (or whatever all of these routines will
!               search on).

        i = 1
        do while (i <= n_fluids)
          if (mfluid_db(i)%str_props(1) == fluid) then
            index = i
            exit
          endif
          i = i + 1
        enddo

!----------------------------------------------------------------------------

        return
        end subroutine find_fluid_index

!****************************************************************************

!  Routine:     check_fluid_index (index, status)

!  Description: Determine whether the specified index is valid.  Returns
!               mfluid_status_fail if it is invalid or the database has
!               not been initialized, otherwise mfluid_status_ok.

        subroutine check_fluid_index (index, status)

!----------------------------------------------------------------------------

        integer :: index, status

!----------------------------------------------------------------------------

        status = mfluid_status_fail

        if (.not. init_was_called) return

        if ((index < 1) .or. (index > n_fluids)) return

        status = mfluid_status_ok

!----------------------------------------------------------------------------

        return
        end subroutine check_fluid_index

!****************************************************************************

!  Routine:     check_prop_name (prop, type, status)

!  Description: Determine whether the specified property name and type
!               match those of a known property.  If they do, return the
!               index of that property, otherwise return mfluid_status_fail.

        subroutine check_prop_name (prop, type, status)

!----------------------------------------------------------------------------

        character(len=*)        :: prop
        integer                 :: type, status

        integer                 :: i

!----------------------------------------------------------------------------

        status = mfluid_status_fail

        i = 1
        do while (i <= n_props)
          if ( (mfluid_props(i)%name == prop) .and.
     &         (mfluid_props(i)%type == type) ) then
            status = mfluid_props(i)%index
            exit
          endif
          i = i + 1
        enddo

!----------------------------------------------------------------------------

        return
        end subroutine check_prop_name

!****************************************************************************

!  Routine:     list_mfluid_db (lun)

!  Description: List the contents of the multifluid database in a
!               snappy table format.  Output goes to the logical I/O
!               unit indicated by the lun parameter.


        subroutine list_mfluid_db (lun)

!----------------------------------------------------------------------------

        integer                 :: lun

        integer                 :: i
        character(len=80)       :: fmt

!----------------------------------------------------------------------------

!               If the database hasn't been initialized, exit.

        if (.not. init_was_called) then
          write (lun,*)
     &      'list_mfluid_db:  init_mfluid_db has not been called'
          return
        endif

        write (lun,*)
        write (lun,*) 'Multifluid database contents:'
        write (lun,*)
        write (lun,"(1X, A9, 11X, 1X, A6, 50(1X, :, A6))")
     &          "Full name", "Abbrev", "A", "Z", "N", "E", "Eb",
     &          "gamma"
        write (lun,"(1X, 79('-'))")

        fmt = "(1X, A20, 1X, A6, 50(1X, :, F6.2))"

        do i = 1, n_fluids
          write (lun,fmt) mfluid_db(i)%str_props,
     &                    mfluid_db(i)%real_props
        enddo

        write (lun,"(1X, 79('-'))")
        write (lun,*)

!----------------------------------------------------------------------------

        return
        end subroutine list_mfluid_db

!****************************************************************************
 
!  Routine:     make_lowercase()
 
!  Description: Convert a string to lowercase.
 
        subroutine make_lowercase (str)
 
        character(len=*):: str
        integer         :: i
 
        do i = 1, len_trim(str)
          if (lge(str(i:i), 'A') .and. lle(str(i:i), 'Z'))
     &      str(i:i) = achar( iachar(str(i:i)) + 32 )
        enddo
 
        return
        end subroutine make_lowercase
 
!****************************************************************************

        end module multifluid_database
